/*! ********************************************
 * SMFSequencer.java
 * Copyright 2019 masahiro takamoto (MotChan)
 * Distributed under the MIT license. 
 * see -> sora/license.txt
 ******************************************** */
package sora.graybox.smf;

import sora.graybox.GrayBox;
import sora.graybox.IGrayBoxModule;
import sora.graybox.MidiEvent;
import sora.graybox.ProcessData;

public final class SMFSequencer implements IGrayBoxModule
{
// EventBuffer -----
	private MidiEvent[] _eventBuffer = null;
	private int _eventBufferSize = 256;
	private int _usedEvents      = 0;
	
	public void clearMIDIEventBuffer()
	{
		for (int i = 0; i < _eventBufferSize; i++)_eventBuffer[i].clear();
		_usedEvents = 0;
	}
	public int getEventBufferSize() { return _eventBufferSize; }
	public int setEventBufferSize(int size)
	{
		if (size == _eventBufferSize) return _eventBufferSize;
		_eventBuffer = new MidiEvent[size];
		for (int i = 0; i < size; i++)_eventBuffer[i] = new MidiEvent();
		_usedEvents = 0;
		return _eventBufferSize = size;
	}
// Sequence -----
	private SMFSequence _sequence = null;
	private SMFTrackEvent _track = null; // Event List - "MTrk" Events
	private int _division = 480;  // "MThd" divison
	private boolean _loopable = false;
	private SMFTrackEvent _loopStart = null;
	private SMFTrackEvent _loopEnd = null;
	public SMFSequence getSequence() { return _sequence; }
	public boolean setSequence(SMFSequence seq)
	{
		initialize();
		if (seq.track == null) return false;
		// ---
		_sequence = seq;
		_track = seq.track;
		_division = seq.division;
		_loopable = seq.loopable;
		_loopStart = seq.loopStart;
		_loopEnd = seq.loopEnd;
		// ---
		_refTime = _division * 1000000.0f / _sampleRate;
		_evTimerStep = (int)(_refTime / _tempo * _EV_TIMER_FPVALUE_N);
		gotoHead();
		return true;
	}
	public boolean loopable() { return _loopable; }
// Sequencer -----
	private int midiOutPorts = 0xFFFF;
	/*
	   if (event.type <= _eventFilter) --> process
	   0xE0 = ChannelEvent Only
	   0xFF = include SysEx
	   0xFFFF = include SysEx and MetaEvent
	*/
	private int _eventFilter = 0xFF;
	private float _sampleRate = 44100.0f;
	private double _tempo = 500000.0;  // Set Tempo (MetaEvent FF51)
	private boolean _isPlay = false;
	private boolean _loopPlay = false;
	private SMFTrackEvent _trackPos = null;
	// event delay timer
	private final int _EV_TIMER_FPBITS  = 24;
	private final int _EV_TIMER_FPVALUE = 1 << _EV_TIMER_FPBITS;
	private final int _EV_TIMER_FPMASK  = _EV_TIMER_FPVALUE - 1;
	private final double _EV_TIMER_FPVALUE_N = (double)_EV_TIMER_FPVALUE;
	private double _refTime = 0.0;  // val = _timeDivision(Note Ticks) * (1000000uSec / SampleRate)
	private int _evTimer    = 0;
	private int _evTimerAFP = 0;
	private int _evTimerStep = 0;
	// getter, setter ---
	public float getSampleRate() { return _sampleRate; }
	public void setSampleRate(float val)
	{ 
		_sampleRate = val;
		_refTime = _division * 1000000.0 / _sampleRate;
		_evTimerStep = (int)(_refTime / _tempo * _EV_TIMER_FPVALUE_N);
	}
	public boolean getOutputMetaMessage() { return _eventFilter == 0xFFFF; }
	public void setOutputMetaMessage(boolean flag) { _eventFilter = (flag)? 0xFFFF : 0xFF; }
	public boolean playable() { return _track != null; }
	public boolean isPlay() { return _isPlay; }
	public boolean isLoopPlay() { return _loopPlay; }
	public boolean endOfTrack() { return _trackPos == null; }
// Sequencer Contorols ---
	public void reset()
	{
		// seq ---
		_tempo = 500000.0;
		_isPlay = false;
		_loopPlay = false;
		_trackPos = null;
		_refTime = 0.0;
		_evTimer = 0;
		_evTimerAFP = 0;
		_evTimerStep = 0;
		gotoHead();
	}
	public void initialize()
	{
		// data ---
		_sequence = null;
		_track = null;
		_division = 480;
		_loopable = false;
		_loopStart = null;
		_loopEnd = null;
		// seq ---
		_tempo = 500000.0;
		_isPlay = false;
		_loopPlay = false;
		_trackPos = null;
		_refTime = 0.0;
		_evTimer = 0;
		_evTimerAFP = 0;
		_evTimerStep = 0;
	}
	public double getBPM() { return 60000000.0 / _tempo; }
	public void setBPM(double val)
	{ 
		_tempo = 60000000.0 / val;
		_evTimerStep = (int)(_refTime / _tempo * _EV_TIMER_FPVALUE_N);
	}
	public boolean play(){ return play(false); }
	public boolean play(boolean resume)
	{
		if (_track == null) return false;
		if (!resume || _trackPos == null)gotoHead();
		return _isPlay = true;
	}
	public void stop(){ stop(false); }
	public void stop(boolean pause)
	{
		_isPlay = false;
		if (_track == null) return;
		if (!pause)gotoHead();
	}
	public boolean loopPlay(boolean flag)
	{
		if (_loopable) return _loopPlay = flag;
		else return false;
	}
	public void gotoHead()
	{
		if (_track == null) return;
		_trackPos = _track;
		// ---
		_tempo = 500000.0;
		_evTimerStep = (int)(_refTime / _tempo * _EV_TIMER_FPVALUE_N);
		// ---
		_evTimer = _trackPos.deltaTime;
		_evTimerAFP = 0;
	}
// Interface - IfGrayBoxModule ---
	public void updateSampleRate(float sampleRate){ setSampleRate(sampleRate); }
	public void updateAudioChannels(int ch){}
	public void updateProcessBlockSize(int size){}
	
	public void resetModule(){ clearMIDIEventBuffer(); reset(); }
	public void initiModule(){ clearMIDIEventBuffer(); initialize(); }
	public void connectModule(GrayBox host)
	{ 
		clearMIDIEventBuffer();
		setSampleRate(host.getSampleRate());
	}
	public void disconnectModule(){ clearMIDIEventBuffer(); }
	public void updateBypassStatus(boolean bypass){ clearMIDIEventBuffer(); }
	
	public void startProcess(){ clearMIDIEventBuffer(); }
	public void stopProcess(){ clearMIDIEventBuffer(); }
	public void suspendProcess(){ clearMIDIEventBuffer(); }
	public void resumeProcess(){ clearMIDIEventBuffer(); }
	public void process(ProcessData procData)
	{
		if (!_isPlay) return;
		_usedEvents = 0;
		for (int i = 0, _sampleFrames = procData.sampleFrames; i < _sampleFrames; i++)
		{
			while (_evTimer <= 0)
			{
			// Arbitrary Loop -->
				if (_trackPos == _loopEnd)
				{
					if (_loopPlay)
					{
						_trackPos = _loopStart;
						//if (callBackLoop != null) callBackLoop();
						//continue;
					}
				}
			// Process Event -->
				if (_trackPos.type <= _eventFilter)
				{
					if (_usedEvents < _eventBufferSize)
					{
						MidiEvent _ev = _eventBuffer[_usedEvents++];
						_ev.next = null;  // clear
						_ev.midiPorts = midiOutPorts;
						_ev.offsetFrames = i;
						_ev.type = _trackPos.type;
						_ev.channel = _trackPos.channel;
						_ev.value0 = _trackPos.value0;
						_ev.value1 = _trackPos.value1;
						_ev.data = _trackPos.data;
					}
				}
			// Tempo -->
				if (_trackPos.type == 0xFF51)
				{
					_tempo = _trackPos.value0;
					_evTimerStep = (int)(_refTime / _tempo * _EV_TIMER_FPVALUE_N);
				}
			// NextEvent, EOT -->
				_trackPos = _trackPos.next;
				if (_trackPos != null)_evTimer += _trackPos.deltaTime;
				else 
				{
					if (_loopPlay) { gotoHead(); continue; }
					else { _isPlay = false; _evTimer = 1; i = _sampleFrames; }
				}
			}
			// Counter -----
			_evTimerAFP += _evTimerStep;
			_evTimer -= _evTimerAFP >> _EV_TIMER_FPBITS;
			_evTimerAFP &= _EV_TIMER_FPMASK;
		}
		// addtoList -->
		MidiEvent _list = procData.midiEventList;
		for (int i = 0; i < _usedEvents; i++)
		{
			MidiEvent _ev = _eventBuffer[i];
			while (_list.next != null)
			{
				if (_ev.offsetFrames >= _list.next.offsetFrames) _list = _list.next;
				else break;
			}
			_ev.next = _list.next;
			_list.next = _ev;
		}
	}
// Constructor ---
	public SMFSequencer(){ this(512); }
	public SMFSequencer(int eventBufferSize)
	{
		_eventBufferSize = eventBufferSize;
		_eventBuffer = new MidiEvent[_eventBufferSize];
		for (int i = 0; i < _eventBufferSize; i++) _eventBuffer[i] = new MidiEvent();
		_usedEvents = 0;
	}
}