/*! ********************************************
 * SampleData.java
 * Copyright 2019 masahiro takamoto (MotChan)
 * Distributed under the MIT license. 
 * see -> sora/license.txt
 ******************************************** */
package sora.graybox.synth;

public class SampleData 
{
	/*
	    |          |      |        |
		|  /\    /\|    /\|  /\    |
		|\/  \  /  |\  /  |\/  \/--|
		|     \/   | \/   |        |
		|          |      |        |SoundOff
		|          |      |toRelease
		|          |      |LoopEnd
		|          |LoopStart
		|KeyOn	
		
		- sampleFrames -
		samples : [L,R][L,R]...
		frames    [ 0 ][ 1 ]...
	 */
	int[] sampleData = null;  // Indexed Sample Data
	int sampleFrames = 0;
	int channelCount = 0;
	private float _sampleRate = 0.0f;
	private float _recPitch   = 0.0f;
	// Synth Param
	int pgSample = 0;  // SampleRate
	int pgPitch  = 0;  // RecPitch
	int pgKey    = 0;  // SampleRate - RecPitch
	int plbkS    = 0;  // PlayBack Start Frame [plbkS >= 0]
	int plbkE    = 0;  // PlayBack End Frame   [plbkE <= sampleFrames]
	int loopE    = 0;  // Loop End Frame       [loopE <= plbkE]
	int loopF    = 0;  // Loop Frames          [val = loopE - lpStart]
	int loopFk   = 0;  // Loop Frames          * after -> keyOff
// getter, setter
	public int getSampleFrames() { return sampleFrames; }
	public int getChannelCount() { return channelCount; }
	public float getSampleRate() { return _sampleRate; }
	public void setSampleRate(float sampleRate)
	{ 
		_sampleRate = sampleRate;
		pgSample = (sampleRate <= 1.0f)? 0 : (int)(Math.log((double)sampleRate) * (SynthData.POW2T_OCTV_RES * 1.442695040888963387) + 0.5);
		pgKey = pgSample - pgPitch;
	}
	public float getRecPitch() { return _recPitch; }
	public void setRecPitch(float pitch)
	{ 
		_recPitch = pitch;
		pgPitch = (pitch <= 1.0f)? 0 : (int)(Math.log((double)pitch) * (SynthData.POW2T_OCTV_RES * 1.442695040888963387) + 0.5);
		pgKey = pgSample - pgPitch;
	}
	public int getPlayBackStartPoint() { return plbkS; }
	public int getPlayBackEndPoint() { return plbkE; }
	public int getLoopStartPoint() { return (loopF > 0)? loopE - loopF : 0; }
	public int getLoopEndPoint() { return (loopF > 0)? loopE : 0; }
	public boolean getLoopable() { return loopF > 0; }
	
	public void clear()
	{
		_sampleRate = 0.0f;
		_recPitch = 0.0f;
		sampleFrames = 0;
		channelCount = 0;
		sampleData = null;
		pgSample = 0;
		pgPitch = 0;
		pgKey = 0;
		plbkS = 0;
		plbkE = 0;
		loopE = 0;
		loopF = 0;
		loopFk = 0;
	}
	public void copy(SampleData from){ copy(from, false); }
	public void copy(SampleData from, boolean deepCopy)
	{
		sampleFrames = from.sampleFrames;
		channelCount = from.channelCount;
		_sampleRate = from._sampleRate;
		_recPitch = from._recPitch;
		// Synth Param
		pgSample = from.pgSample;
		pgPitch = from.pgPitch;
		pgKey = from.pgKey;
		plbkS = from.plbkS;
		plbkE = from.plbkE;
		loopE = from.loopE;
		loopF = from.loopF;
		loopFk = from.loopFk;
		// wave
		if (deepCopy && from.sampleData != null)
		{
			int i, _len = from.sampleData.length;
			int[] _source = from.sampleData;
			sampleData = new int[_len];
			for (i = 0; i < _len; i++) sampleData[i] = _source[i];
		}
		else sampleData = from.sampleData;
	}
	public void setPlaybackRange(int start, int end)
	{
		if (start >= end || end > sampleFrames) return;
		plbkS = start;
		plbkE = end;
		loopE = plbkE;
		loopF = 0;
		loopFk = 0;
	}
	public void resetLoopPoints(){ resetLoopPoints(0, 0); }
	public void resetLoopPoints(int start, int end)
	{
		// loop
		if (start >= plbkS && start < end && end <= plbkE)
		{
			loopE = end;
			loopF = end - start;
			loopFk = (end == plbkE)? loopF : 0;
		}
		// no loop
		else
		{
			loopE = plbkE;
			loopF = 0;
			loopFk = 0;
		}
	}
	public void setPCMSamples(float[] wave, int ch)
	{
		clear();
		// ---
		final double _LOGE = -(SynthData.LOG2VT_DYN_RES * 1.442695040888963387);
		final float  _BOADER = 1.0f / (1 << SynthData.LOG2VT_DYN_BITS);
		final int _BOTTOM = SynthData.LOG2VT_SIZE;
		int _len = wave.length;
		int[] _wave = new int[_len];
		for (int i = 0; i < _len; i++)
		{
			float v = wave[i];
			// Positive Value
			if      (v >  _BOADER) _wave[i] =  (int)(Math.log((double) v) * _LOGE + 0.5) << 1;
			// Negative Value
			else if (v < -_BOADER) _wave[i] = ((int)(Math.log((double)-v) * _LOGE + 0.5) << 1) + 1;  // +1 = Nega
			// 0 value
			else _wave[i] = _BOTTOM;
		}
		sampleData = _wave;
		channelCount = ch;
		sampleFrames = _len / ch;
		// pitch
		_sampleRate = 44100.0f;
		setSampleRate(_sampleRate);
		// playback
		plbkE = sampleFrames;
		loopE = sampleFrames;
	}
	public void loadPCMData(byte[] dat, int position, int sampleFrames, int ch, int bitRate, boolean sign, boolean bigEndian)
	{
		clear();
		// ---
		int p = position;
		int _len = sampleFrames * ch;
		int[] _wave = null;
		double _sl;
		int _p0, _p1, _p2, _p3;
		switch (bitRate)
		{
			default: clear(); return;
			
			case  8:
				_wave = new int[_len];
				_sl = 1.0 / 128.0;
				if (sign) for (int i = 0; i < _len; i++) _wave[i] = _log2VTindex(dat[p++] * _sl);
				else
				{
					for (int i = 0; i < _len; i++)
					{
						double _v = (dat[p++] & 0xff) - 128.0;
						_wave[i] = _log2VTindex(_v * _sl);
					}
				}
				break;
				
			case 16:
				_wave = new int[_len];
				_sl = 1.0 / 32768.0;
				if (bigEndian){ _p0 = 0; _p1 = 1; } else { _p0 = 1; _p1 = 0; }
				if (sign)
				{
					for (int i = 0; i < _len; i++, p += 2)
					{
						short _v = (short)(dat[p + _p0] << 8 | (dat[p + _p1]&0xff));
						_wave[i] = _log2VTindex(_v * _sl);
					}
				}
				else
				{
					for (int i = 0; i < _len; i++, p+=2)
					{
						double _v = (int)((dat[p + _p0]&0xff) << 8 | (dat[p + _p1]&0xff)) - 32768.0;
						_wave[i] = _log2VTindex(_v * _sl);
					}
				}
				break;
				
			case 24:
				_wave = new int[_len];
				_sl = 1.0 / 8388608.0;
				if (bigEndian){ _p0 = 0; _p1 = 1; _p2 = 2; }
				else          { _p0 = 2; _p1 = 1; _p2 = 0; }
				if (sign)
				{
					for (int i = 0; i < _len; i++, p += 3)
					{
						int _v = (int)(dat[p + _p0] << 24 | (dat[p + _p1]&0xff) << 16 | (dat[p + _p2]&0xff) << 8) / 256;// >> 8;
						_wave[i] = _log2VTindex(_v * _sl);
					}
				}
				else
				{
					for (int i = 0; i < _len; i++, p+=3)
					{
						double _v = (int)((dat[p + _p0]&0xff) << 16 | (dat[p + _p1]&0xff) << 8 | (dat[p + _p2]&0xff)) - 8388608.0;
						_wave[i] = _log2VTindex(_v * _sl);
					}
				}
				break;
				
			case 32:
				_wave = new int[_len];
				_sl = 1.0 / 2147483648.0;
				if (bigEndian){ _p0 = 0; _p1 = 1; _p2 = 2; _p3 = 3; }
				else          { _p0 = 3; _p1 = 2; _p2 = 1; _p3 = 0; }
				if (sign)
				{
					for (int i = 0; i < _len; i++, p+=4)
					{
						int _v = (int)(dat[p + _p0] << 24 | (dat[p + _p1]&0xff) << 16 | (dat[p + _p2]&0xff) << 8 | (dat[p + _p3]&0xff));
						_wave[i] = _log2VTindex(_v * _sl);
					}
				}
				else
				{
					for (int i = 0; i < _len; i++, p+=4)
					{
						double _v = (long)((dat[p + _p0]&0xff) << 24 | (dat[p + _p1]&0xff) << 16 | (dat[p + _p2]&0xff) << 8 | (dat[p+_p3]&0xff)) - 2147483648.0;
						_wave[i] = _log2VTindex(_v * _sl);
					}
				}
				break;
		}
		sampleData = _wave;
		channelCount = ch;
		this.sampleFrames = sampleFrames;
		// pitch
		_sampleRate = 44100.0f;
		setSampleRate(_sampleRate);
		// playback
		plbkE = sampleFrames;
		loopE = sampleFrames;
	}

	static final double _LOGE = -(SynthData.LOG2VT_DYN_RES * 1.442695040888963387);
	static final double _BOADER = 1.0f / (1 << SynthData.LOG2VT_DYN_BITS);
	static final int _BOTTOM = SynthData.LOG2VT_SIZE;
	static private int _log2VTindex(double val)
	{
		if      (val >  _BOADER) return  (int)(Math.log( val) * _LOGE + 0.5) << 1;
		else if (val < -_BOADER) return ((int)(Math.log(-val) * _LOGE + 0.5) << 1) + 1;  // +1 = Nega
		else return _BOTTOM;
	}

// Constructor
	public SampleData(){}
}