/*! ********************************************
 * SynthData.java
 * Copyright 2019 masahiro takamoto (MotChan)
 * Distributed under the MIT license. 
 * see -> sora/license.txt
 ******************************************** */
package sora.graybox.synth;

public final class SynthData
{
// Instance ---
	static private SynthData _instance = null;
	static SynthData getInstance(){ return (_instance != null)? _instance : (_instance = new SynthData()); }
	static public void disposeInstance(){ _instance = null; }
	private SynthData(){}
// const ---
	// Volume, Pan
	// Log2 Volume Table
	static final int LOG2VT_DYN_BITS     = 16;                    // const 16    96dB dynamic range (FM TL, 16Bit PCM)
	static final int LOG2VT_DYN          = 1 << LOG2VT_DYN_BITS;  //             1bit per 6dBv
	static final int LOG2VT_DYN_RES_BITS = 8;
	static final int LOG2VT_DYN_RES      = 1 << LOG2VT_DYN_RES_BITS;
	static final int LOG2VT_SIZE         = LOG2VT_DYN_BITS << LOG2VT_DYN_RES_BITS + 1;
	// Pow2 Table
	static final int POW2T_OCTV_BITS = 13;  // const 13
	static final int POW2T_OCTV_RES  = 1 << POW2T_OCTV_BITS;  // POW2T_OCTV_RES == LOG2VT_SIZE
	static final int POW2T_OCTV_MASK = POW2T_OCTV_RES - 1;
	// Envelope Generator
	static final int EG_MAX_LEVEL_BITS = 4 + LOG2VT_DYN_RES_BITS;  // constant 12 (4 = 16 LOG2VT_DYN_BITS)
	static final int EG_MAX_LEVEL      = 1 << EG_MAX_LEVEL_BITS;   // 4096 = 16 * 256
	static final int EG_MAX_CNT_BITS   = 30;
	static final int EG_MAX_CNT        = 1 << EG_MAX_CNT_BITS;
	static final int EG_CNT_FP         = EG_MAX_CNT_BITS - EG_MAX_LEVEL_BITS;  // for EGLevelTable.  xbit fixed-point  (precision 1 / (1 << x))
	static final int EG_ENV_DYN_BITS = 13;  // 1 ~ 16
	static final int EG_ENV_DYN      = 1 << EG_ENV_DYN_BITS;
	static final int EG_CNT_ENV_ATK  = (int)(1.0 / EG_ENV_DYN * EG_MAX_LEVEL) << EG_CNT_FP;  // 0
	static final int EG_CNT_ENV_BTM  = (EG_ENV_DYN_BITS * LOG2VT_DYN_RES) << EG_CNT_FP;
	static final int EG_SUSTAIN_BTM  = EG_ENV_DYN_BITS * 8;
	static final double EG_REFERENCE_TIME = 16.0;                             // to 1/2 Volt (-6dB) Time [EG : Rate  0]
	static final double EG_SOUND_OFF_TIME = 1.0 / (EG_REFERENCE_TIME * 2.0);  // to 1/2 Volt (-6dB) Time [EG : Rate 34]
	// LFO
	static final double LFO_MAX_FREQ   = 32.0;  // Hz
	static final int LFO_TIMER_INIT_VALUE     = 0x40000000;  // initial Value - Phase Update Timer
	static final int LFO_WAVE_TABLE_SIZE_BITS = 8;
	static final int LFO_WAVE_TABLE_SIZE      = 1 << LFO_WAVE_TABLE_SIZE_BITS;  // PM[0,2,4...],AM[1,3,5...]
	// Phase Modulation
// Tables ---
	// Volume, Pan
	private float[] _VOLUME_TABLE_0dB = null;
	private float[] _VOLUME_TABLE_6dB = null;
	private float[] _PAN_TABLE_SIN    = null;
	private float[] _PAN_TABLE_BAL    = null;
	// Log2 Volume Table
	private float[] _LOG2V_TABLE = null;
	private int[] _LOG2V_TABLE_I = null;
	// Pow2 Table
	private int[] _POW2_TABLE      = null;
	private int[] _POW2T_RS_TABLE  = null;
	private int[] _POW2T_KEY_TABLE = null;
	// Envelope Generator
	private int[] _EG_TOTAL_LEVEL_TABLE = null;
	private int[] _EG_LEVEL_TABLE_UP    = null;
	private int[] _EG_LEVEL_TABLE_DW    = null;
	private int[] _EG_KEY_TABLE      = null;
	private int[] _EG_POW2T_RS_TABLE = null;
	private int[] _EG_CNT_GATE_TABLE = null;
	// LFO
	private int[][] _LFO_WAVE_TABLES = null;
	// Phase Modulation
	private int[] _MUL_TABLE = null;

// getter, creator ---
// Volume, Pan ==================================================
	static float[] createVOLUME_TABLE(int size, double c, boolean reverse)
	{
		// Curve ( quadratic x^2 )
		// Fader |    dB | volt
		//  100% | - 0db | 1.0
		//   50% | -12db | 0.25
		//   25% | -24db | 0.0625
		float[] _table = new float[size];
		int j = 1, _cnt = 1;
		double _sizeD = (double)size;
		if (reverse) { j = size; _cnt = -1; }
		for (int i = 0; i < size; i++, j+=_cnt)_table[i] = (float)(Math.pow(j / _sizeD, 2.0) * c);
		return _table;
	}
	private float[] _createPAN_TABLE(int type)
	{
		final double _PIS2 = 3.141592653589793 / 2.0;
		float[] _table = new float[257];
		switch (type)
		{
			// Panning Law (Balance Pan)
			// Fader         |    dB | volt
			//  100%         | - 0db | 1.0
			//   50%(Center) | - 0db | 1.0
			//   25%         | - 6db | 0.5
			case 1:
				for (int i = 0; i < 64; i++)_table[i] = 1.0f;
				for (int i = 64, v = 0; i < 192; i++, v++)_table[i] = 1.0f - v / 64.0f;
				for (int i = 192; i < 257; i++)_table[i] = 0.0f;
				break;
			// Panning Law (Combine Pan)
			// Fader         |    dB | volt
			//  100%         | - 0db | 1.0
			//   50%(Center) | - 6db | 0.5
			//   25%         | -12db | 0.25
			case 2:
				for (int i = 0; i < 64; i++)_table[i] = 1.0f;
				for (int i = 64, v = 0; i < 192; i++, v++)_table[i] = (128-v) / 128.0f;
				for (int i = 192; i < 257; i++)_table[i] = 0.0f;
				break;
			// Panning Law (Sine Curve)
			default:
				for (int i = 0; i < 64; i++)_table[i] = 1.0f;
				for (int i = 64, v = 0; i < 192; i++, v++)_table[i] = (float)Math.sin(((128 - v) / 128.0) * _PIS2);
				for (int i = 192; i < 257; i++)_table[i] = 0.0f;
				break;
		}
		return _table;
	}
	float[] getVOLUME_TABLE_0dB(){ return (_VOLUME_TABLE_0dB != null)? _VOLUME_TABLE_0dB : (_VOLUME_TABLE_0dB = createVOLUME_TABLE(128, 1.0, false)); } 
	float[] getVOLUME_TABLE_6dB(){ return (_VOLUME_TABLE_6dB != null)? _VOLUME_TABLE_6dB : (_VOLUME_TABLE_6dB = createVOLUME_TABLE(128, 0.5011872336272722, false)); } 
	float[] getPAN_TABLE_SIN(){ return (_PAN_TABLE_SIN != null)? _PAN_TABLE_SIN : (_PAN_TABLE_SIN = _createPAN_TABLE(0)); } 
	float[] getPAN_TABLE_BAL(){ return (_PAN_TABLE_BAL != null)? _PAN_TABLE_BAL : (_PAN_TABLE_BAL = _createPAN_TABLE(1)); } 
// Log2 Volume Table ==================================================
	static private final double _LOG2VT_LOGE =  -(LOG2VT_DYN_RES * 1.442695040888963387); // -ln(2)
	static private final float  _LOG2VT_BOADERf = 1.0f / LOG2VT_DYN;
	static private final double _LOG2VT_BOADERd = 1.0 / LOG2VT_DYN;
	// in [-1.0 ~ 1.0]
	static int toLog2vtIndexValue(float vol)
	{
		// Positive Value
		if      (vol >  _LOG2VT_BOADERf) return  (int)(Math.log((double) vol) * _LOG2VT_LOGE + 0.5) << 1;
		// Negative Value
		else if (vol < -_LOG2VT_BOADERf) return ((int)(Math.log((double)-vol) * _LOG2VT_LOGE + 0.5) << 1) + 1;  // +1 = Nega
		// 0 value
		else return LOG2VT_SIZE;
	}
	static int toLog2vtIndexValue(double vol)
	{
		// Positive Value
		if      (vol >  _LOG2VT_BOADERd) return  (int)(Math.log( vol) * _LOG2VT_LOGE + 0.5) << 1;
		// Negative Value
		else if (vol < -_LOG2VT_BOADERd) return ((int)(Math.log(-vol) * _LOG2VT_LOGE + 0.5) << 1) + 1;  // +1 = Nega
		// 0 value
		else return LOG2VT_SIZE;
	}
	float[] getLOG2V_TABLE()
	{
		if (_LOG2V_TABLE != null) return _LOG2V_TABLE;
		final double _dynB = (double)LOG2VT_DYN_BITS, _maxV = (double)LOG2VT_DYN, _res = (double)LOG2VT_DYN_RES;
		final int _tableSize = (LOG2VT_SIZE + 2) * 4;  // Wave + EG(TL) + EG(DYN) + AM (+2 = safe)
		float[] _table = new float[_tableSize];
		for (int i = 0, j = 0; i < LOG2VT_SIZE; j++)
		{
			float _nv = (float)(Math.pow(2.0, _dynB - j / _res) / _maxV);
			_table[i++] =  _nv;
			_table[i++] = -_nv;
		}
		for (int i = LOG2VT_SIZE; i < _tableSize; ++i)_table[i] = 0.0f;
		return _LOG2V_TABLE = _table;
	}
	int[] getLOG2V_TABLE_I()
	{
		if (_LOG2V_TABLE_I != null) return _LOG2V_TABLE_I;
		final double _dynB = (double)LOG2VT_DYN_BITS, _res = (double)LOG2VT_DYN_RES;
		final int _tableSize = (LOG2VT_SIZE + 2) * 4;  // Wave + EG(TL) + EG(DYN) + AM (+2 = safe)
		int[] _table = new int[_tableSize];
		for (int i = 0, j = 0; i < LOG2VT_SIZE; j++)
		{
			int _iv = (int)(Math.pow(2.0, _dynB - j / _res));
			_table[i++] =  _iv;
			_table[i++] = -_iv;
		}
		for (int i = LOG2VT_SIZE; i < _tableSize; ++i)_table[i] = 0;
		return _LOG2V_TABLE_I = _table;
	}
// Pow2 Table ==================================================
	static private final double _POW2T_LOGE = POW2T_OCTV_RES * 1.442695040888963387;
	// in [1.0 ~ inf]
	static int toPow2tIndexValue(float val)
	{
		if (val <= 1.0f) return 0;
		else return (int)(Math.log((double)val) * _POW2T_LOGE + 0.5);
	}
	static int toPow2tIndexValue(double val)
	{
		if (val <= 1.0) return 0;
		else return (int)(Math.log(val) * _POW2T_LOGE + 0.5);
	}
	int[] getPOW2_TABLE()
	{
		if (_POW2_TABLE != null) return _POW2_TABLE;
		final double _octvRes = (double)POW2T_OCTV_RES;
		int[] _table = new int[POW2T_OCTV_RES];
		for (int i = 0; i < POW2T_OCTV_RES; i++)_table[i] = (int)(Math.pow(2.0, 30.0 + (i / _octvRes)) + 0.5);
		return _POW2_TABLE = _table;
	}
	int[] getPOW2T_RS_TABLE()
	{
		if (_POW2T_RS_TABLE != null) return _POW2T_RS_TABLE;
		int[] _table = new int[128];
		for (int i =  0; i <  30; i++)_table[i] = 30 - i;
		for (int i = 30; i < 128; i++)_table[i] = 0;  // Safe Field
		return _POW2T_RS_TABLE = _table;
	}
	int[] getPOW2T_KEY_TABLE()
	{
		if (_POW2T_KEY_TABLE != null) return _POW2T_KEY_TABLE;
		final double _octvRes = (double)POW2T_OCTV_RES;
		int[] _table = new int[128];
		for (int i = 0; i < 128; i++)_table[i] = (int)(i / 12.0 * _octvRes + 0.5);
		return _POW2T_KEY_TABLE = _table;
	}
// Envelope Generator ==================================================
	int[] getEG_TOTAL_LEVEL_TABLE()
	{
		if (_EG_TOTAL_LEVEL_TABLE != null) return _EG_TOTAL_LEVEL_TABLE;
		// Ref - YM2151
			// D6 D5 D4 D3 D2 D1  D0
			// 48 24 12 6  3  1.5 0.75 [dB]
		final int _step = LOG2VT_SIZE >> 7;
		int[] _table = new int[128*2];
		for (int i =   0; i < 128; i++)_table[i] = i * _step;  // 64(0.75dB) = 8192/128
		for (int i = 128; i < 256; i++)_table[i] = LOG2VT_SIZE;  // VeloCtrl
		return _EG_TOTAL_LEVEL_TABLE = _table;
	}
	int[] getEG_LEVEL_TABLE_DW()
	{
		if (_EG_LEVEL_TABLE_DW != null) return _EG_LEVEL_TABLE_DW;
		int[] _table = new int[EG_MAX_LEVEL];
		for (int i = 0; i < EG_MAX_LEVEL; i++)_table[i] = i << 1;
		return _EG_LEVEL_TABLE_DW = _table;
	}
	int[] getEG_LEVEL_TABLE_UP()
	{
		if (_EG_LEVEL_TABLE_UP != null) return _EG_LEVEL_TABLE_UP;
		final double _size = (double)EG_MAX_LEVEL;
		int[] _table = new int[EG_MAX_LEVEL];
		for (int i = 0; i < EG_MAX_LEVEL; i++)_table[i] = toLog2vtIndexValue(i / _size);
		return _EG_LEVEL_TABLE_UP = _table;
	}
	int[] getEG_KEY_TABLE()
	{
		if (_EG_KEY_TABLE != null) return _EG_KEY_TABLE;
		final double _octvRes = (double)POW2T_OCTV_RES;
		int[] _table = new int[256];
		for (int i = 0; i < 128; i++)_table[i] = 0;
		for (int i = 128, _b = 1; i < 256; i++, _b++)_table[i] = (int)(_b / 48.0 * _octvRes);
		return _EG_KEY_TABLE = _table;
	}
	int[] getEG_POW2T_RS_TABLE()
	{
		if (_EG_POW2T_RS_TABLE != null) return _EG_POW2T_RS_TABLE;
		int[] _table = new int[128];
		for (int i =  0; i <= EG_MAX_CNT_BITS; i++)_table[i] = 30 - i;
		for (int i = EG_MAX_CNT_BITS; i < 128; i++)_table[i] = 31 - EG_MAX_CNT_BITS; // Safe Field (value >= 1)
		return _EG_POW2T_RS_TABLE = _table;
	}
	int[] getEG_CNT_GATE_TABLE()
	{
		if (_EG_CNT_GATE_TABLE != null) return _EG_CNT_GATE_TABLE;
			// D6 D5 D4 D3 D2 D1  D0
			// 48 24 12 6  3  1.5 0.75 [dB]
		final int _btm = EG_ENV_DYN_BITS * 8, 
		          _step = (EG_MAX_LEVEL << EG_CNT_FP) / 128;
		int[] _table = new int[129];
		for (int i =    0; i < _btm; i++)_table[i] = i * _step;
		for (int i = _btm; i <  129; i++)_table[i] = EG_CNT_ENV_BTM;
		return _EG_CNT_GATE_TABLE = _table;
	}
// LFO ==================================================
	static private float LFO_MAX_FREQf = (float)LFO_MAX_FREQ;
	static int lfoTimerCoef(float sampleRate, int steps)
	{
		float _tableSize = LFO_WAVE_TABLE_SIZE / 2.0f,
		// Max Frame Rate - LFO Phase Step
		       _lfoMaxFrameRate = sampleRate / LFO_MAX_FREQf / _tableSize;
		if (_lfoMaxFrameRate < 1.0f)_lfoMaxFrameRate = 1.0f;  // Max 1frame per 1Phase
		return (int)(LFO_TIMER_INIT_VALUE / steps / _lfoMaxFrameRate + 0.5f);
	}
	static int lfoTimerCoef(double sampleRate, int steps)
	{
		double _tableSize = LFO_WAVE_TABLE_SIZE / 2.0,
		// Max Frame Rate - LFO Phase Step
		       _lfoMaxFrameRate = sampleRate / LFO_MAX_FREQ / _tableSize;
		if (_lfoMaxFrameRate < 1.0)_lfoMaxFrameRate = 1.0;  // Max 1frame per 1Phase
		return (int)(LFO_TIMER_INIT_VALUE / steps / _lfoMaxFrameRate + 0.5);
	}
	int[][] getLFO_WAVE_TABLES()
	{
		if (_LFO_WAVE_TABLES != null) return _LFO_WAVE_TABLES;
		final int _SIZ = LFO_WAVE_TABLE_SIZE, _HLF = _SIZ >> 1,
		      _QTR = _SIZ >> 2, _TRI = _HLF + _QTR;
		final double _SIZ_N = _SIZ, _HLF_N = _SIZ_N / 2.0, _QTR_N = _SIZ_N / 4.0;
		final double _LOG2VT_SIZEd = (double)LOG2VT_SIZE, _POW2T_OCTV_RESd = (double)POW2T_OCTV_RES;
		int[] _table = null;
		int[][] _tables = new int[4][];
	// Tri ---
		_tables[0] = _table = new int[_SIZ];
		// PM
		for (int i = 0; i < _QTR; i += 2)
		{
			int _iv = (int)(i / _QTR_N * _POW2T_OCTV_RESd);
			_table[       i] = _iv;
			_table[_QTR + i] = POW2T_OCTV_RES - _iv;
			_table[_HLF + i] = -_iv;
			_table[_TRI + i] = _iv - POW2T_OCTV_RES;
		}
		// AM
		for (int i = 0; i < _HLF; i+=2)	
		{
			int _iv = (int)(i / _HLF_N * _LOG2VT_SIZEd);
			_table[       i + 1] = _iv;
			_table[_HLF + i + 1] = LOG2VT_SIZE - _iv;
		}
	// Rect ---
		_tables[1] = _table = new int[_SIZ];
		for (int i = 0; i < _HLF; i+=2)	
		{
			// PM	
			_table[       i] =  POW2T_OCTV_RES;
			_table[_HLF + i] = -POW2T_OCTV_RES;
			// AM
			_table[       i + 1] = 0;
			_table[_HLF + i + 1] = LOG2VT_SIZE;
		}
	// Saw ---
		_tables[2] = _table = new int[_SIZ];
		for (int i = 0; i < _SIZ; i+=2)	
		{
			// PM	
			_table[i    ] = POW2T_OCTV_RES - (int)(i / _HLF_N * _POW2T_OCTV_RESd);
			// AM
			_table[i + 1] = (int)(i / _SIZ_N * _LOG2VT_SIZEd);
		}
	// Noise ---
		_tables[3] = _table = new int[_SIZ];
		for (int i = 0; i < _SIZ; i+=2)	
		{
			double _nv = Math.random();
			// AM
			_table[i + 1] = (int)(_nv * _LOG2VT_SIZEd);
			// PM
			_nv = _nv * 2.0 - 1.0;
			_table[i    ] = (int)(_nv * _POW2T_OCTV_RESd);
		}
		return _LFO_WAVE_TABLES = _tables;
	}
// Phase Modulation ==================================================
	int[] getMUL_TABLE()
	{
		if (_MUL_TABLE != null) return _MUL_TABLE;
		final int _octvRes = POW2T_OCTV_RES; 
		int[] _table = new int[16];
		for (int i = 0; i < 16; i++)_table[i] = (i-1) * _octvRes;
		return _MUL_TABLE = _table;
	}
}