/*! ********************************************
 * SineGenerator.java
 * Copyright 2022 masahiro takamoto (MotChan)
 * Distributed under the MIT license. 
 * see -> license.txt
 ******************************************** */
import sora.graybox.GrayBox;
import sora.graybox.IGrayBoxModule;
import sora.graybox.MidiEvent;
import sora.graybox.ProcessData;

public class SineGenerator implements IGrayBoxModule {
	public float volume = 0.1f;  // 音量
	private boolean _keyOn = false;  // 出音フラグ
	private double _sampleRate = 44100.0;  // サンプリング周波数
	private double _frequency = 440.0;  // 出音周波数
	private double _af = _frequency * 6.28318530718;  // 角周波数
	private int _phase = 0;  // 位相 (周期 = サンプリング周波数)
	// コンストラクター
	public SineGenerator() {}
	
// sora.graybox.IGrayBoxModule -->
	// ホストのサンプリングレートが変更された時に呼ばれるメソッド
	// GrayBox.setSampleRate() -> このメソッド
	public void updateSampleRate(float sampleRate) {
		double n = _phase / _sampleRate;
		n = (int)n - n;
		_sampleRate = sampleRate;
		_phase = (int)(n * sampleRate + 0.5);
	}
	// ホストのオーディオチャンネル数が変更された時に呼ばれるメソッド
	// GrayBox.setAudioChannels() -> このメソッド
	public void updateAudioChannels(int ch){}
	// ホストのプロセスブロックサイズが変更された時に呼ばれるメソッド
	// GrayBox.setProcessBlockSize() -> このメソッド
	public void updateProcessBlockSize(int size){}
	// ホストのGrayBox.resetModules()メソッドが使用された時に呼ばれるメソッド
	// GrayBox.resetModules() -> このメソッド
	public void resetModule(){ volume = 0.1f; _keyOn = false; }
	// ホストのGrayBox.initiModules()メソッドが使用された時に呼ばれるメソッド
	// GrayBox.initiModules() -> このメソッド
	public void initiModule(){ volume = 0.1f; _keyOn = false; }
	// このモジュールがホストへ接続された時に呼ばれるメソッド
	// GrayBox.attachModule() -> このメソッド
	public void connectModule(GrayBox host) {
		_keyOn = false;
		updateSampleRate(host.getSampleRate());
	}
	// このモジュールがホストから取り外された時に呼ばれるメソッド
	// GrayBox.detachModule() -> このメソッド
	public void disconnectModule(){}
	// このモジュールが接続されているスロットの、バイパス状況が更新された時に呼ばれるメソッド
	// GrayBox.bypassProcess() -> このメソッド
	public void updateBypassStatus(boolean bypass){ if (bypass)_keyOn = false; }
	// ホストのGrayBox.startProcess()メソッドが使用された時に呼ばれるメソッド
	// GrayBox.startProcess() -> このメソッド
	public void startProcess(){ _keyOn = false; }
	// ホストのGrayBox.stopProcess()メソッドが使用された時に呼ばれるメソッド
	// GrayBox.stopProcess() -> このメソッド
	public void stopProcess(){ _keyOn = false; }
	// ホストのGrayBox.suspendProcess()メソッドが使用された時に呼ばれるメソッド
	// GrayBox.suspendProcess() -> このメソッド
	public void suspendProcess(){}
	// ホストのGrayBox.resumeProcess()メソッドが使用された時に呼ばれるメソッド
	// GrayBox.resumeProcess() -> このメソッド
	public void resumeProcess(){}
	// ホストが音声処理を実行した時に呼ばれるメソッド
	// GrayBox.process() -> このメソッド
	public void process(ProcessData procData) {
	// 音声処理準備 -->
		int _startFrame = procData.offsetFrames;  // 音声処理開始フレーム
		int _endFrame = _startFrame + procData.sampleFrames;  // 終了フレーム
	// 音声処理、MIDIイベント処理 -->
		// GrayBoxシステムにおけるMIDIメッセージの伝送はProcessDataオブジェクトが持つ
		// MIDIイベントリスト(片方向リンクリスト)を使用して行われます。
		// MIDIメッセージの送信側はこのリストにMIDIイベントを登録していき、
		// MIDIメッセージの受信側はこのリストからMIDIイベント参照してメッセージの処理を行います。
		
		// ・ProcessData.midiEventList
		//   MIDIイベントリスト(先頭ノード(リスト管理用ダミーノード))
		//   このリストのノードの並びはMidiEvent.offsetFramesの昇順です。
		// ・ProcessData.getMidiEvents() : MidiEvent
		//   リスト先頭のダミーノードを除くリストを取得(通常はこちらを使用してください)
		//   リストに何も登録されていなければnullが返ります。
		MidiEvent _event = procData.getMidiEvents();
		// リストからMIDIイベントを順に読み込み、処理を行う
		while (_event != null) {
			// 出音フラグが立っていれば、このMIDIイベントまで音声処理
			// MidiEvent.offsetFrames -> ブロック先頭からのフレームオフセット
			if (_keyOn)_sineGen(procData.audioOut, _startFrame, _event.offsetFrames);
			
			// MIDIメッセージタイプ別で処理の振り分け
			switch (_event.type) {
				case 0x80: // ノートオフ
					_startFrame = _event.offsetFrames;  // 開始フレーム更新
					_keyOn = false;  // 出音フラグ オン
					break;
					
				case 0x90: // ノートオン
					_startFrame = _event.offsetFrames; // 開始フレーム更新
					_keyOn = true;  // 出音フラグ オン
					// 出音条件の更新
					_frequency = 440.0 * Math.pow(2.0, (_event.value0 - 69.0) / 12.0); // 周波数更新
					_af = _frequency * 6.28318530718; // 角周波数更新
					_phase = 0;  // 位相の初期化
					break;
			}
			// 次のMIDIイベント
			_event = _event.next;
		}
		// 終了フレームまで音声処理
		if (_keyOn)_sineGen(procData.audioOut, _startFrame, _endFrame);
	}
	private void _sineGen(float[][] out, int startFrame, int frameEnd) {
		float[] _outL = out[0];
		float[] _outR = out[1];
		for (int i = startFrame; i < frameEnd; i++) {
			float n = (float)Math.sin((_phase / _sampleRate) * _af) * volume; // サイン波出力
			_phase++;  // 位相の更新
			// オーディオバッファに対して加算代入
			_outL[i] += n;
			_outR[i] += n;
			// このモジュールしか音声処理を行わない場合は代入だけでも良い
			/* 
				_outL[i] = _n; 
				_outR[i] = _n;
			*/
		}
	}
}
